#include "StdAfx.h"
#include <typeinfo>

#include "WaterSurface.h"
#include "Engine.h"

const int CUBE_MAP_SIZE = 512;

Engine* Engine::instance = 0;

Engine::Engine(void)
: paused(false), unpaused(false), framebyframe(false), fbf_advance(false), firstframe(true), fbf_ticks(5), 
moved(true), aviMode(false), aviFrames(300), aviFPS(15), ttTotal("Total time")
{
}

Engine::~Engine(void)
{
	for (std::list<IDrawable*>::iterator it = drawlist.begin(); it != drawlist.end(); it++)
	{
		delete (*it);
	}
}

Engine* Engine::Instance()
{
	if (instance == 0)
	{
		instance = new Engine();
	}

	return instance;
}

void Engine::Initialize(int width, int height, int surfaceSize)
{
	vpWidth = width;
	vpHeight = height;
	size = surfaceSize;

	glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);
	glutInitWindowSize(vpWidth, vpHeight);
	glutInitWindowPosition(0, 0);

	glutCreateWindow("The Well");

	glewInit();
	if (GLEW_VERSION_2_0) {
		std::cout << "Ready for OpenGL 2.0" << std::endl;
	} else {
		std::cout << "OpenGL 2.0 not supported" << std::endl;
		exit(1);
	}

	/* Initialize DevIL */
	ilInit();
	if(ilGetError() != IL_NO_ERROR) exit(-1);
	ilutRenderer(ILUT_OPENGL);
	if(ilGetError() != IL_NO_ERROR) exit(-1);

	GLfloat ambient_light[] = { 0.2f, 0.2f, 0.2f, 1.0f };
	GLfloat light_position[] = { 0.0f, 150.0f, 40.0f, 1.0f };

	glEnable(GL_DEPTH_TEST);

	glGenTextures(1, &texCubeMap);
	// bind cube map
	glEnable(GL_TEXTURE_CUBE_MAP);
    glBindTexture(GL_TEXTURE_CUBE_MAP, texCubeMap);
    // build texture (all 6 faces)
    glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	for (int i = 0; i < 6; ++i) {
		glTexImage2D(GL_TEXTURE_CUBE_MAP_POSITIVE_X + i, 0, GL_RGB, CUBE_MAP_SIZE, CUBE_MAP_SIZE, 0, GL_RGB, GL_UNSIGNED_BYTE, 0);
    }

	glClearColor(1.0f, 1.0f, 1.0f, 1.0f);

	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);
	glLightfv(GL_LIGHT0, GL_POSITION, light_position);
	glLightModelfv(GL_LIGHT_MODEL_AMBIENT, ambient_light);
	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
	
	eye_coord[0] = 64.0f; eye_coord[1] = 40.0f; eye_coord[2] = -20.0f;
	line_of_sight[0] = 0.0f; line_of_sight[1] = -35.0f; line_of_sight[2] = 127.0f;

	glCullFace(GL_BACK);
	glEnable(GL_CULL_FACE);
	glShadeModel(GL_SMOOTH);

	glutDisplayFunc(gcDisplay);
	glutReshapeFunc(gcReshape);
	glutIdleFunc(gcIdle);
	glutKeyboardFunc(gcKeyboard);
	glutSpecialFunc(gcSpecial);

	wglSwapIntervalEXT(0);
}

void Engine::Add(IDrawable *object)
{
	drawlist.push_back(dynamic_cast<IDrawable*>(object));

	if (dynamic_cast<IKeyEvent*>(object))
	{
		keylist.push_back(dynamic_cast<IKeyEvent*>(object));
	}
	if (dynamic_cast<IStatistics*>(object))
	{
		statlist.push_back(dynamic_cast<IStatistics*>(object));
	}
}

void Engine::Run()
{
	ttTotal.Start();
	glutMainLoop();
}

void Engine::render()
{
	// compute FPS and elapsed time
	static int oldTime = glutGet(GLUT_ELAPSED_TIME);
	static int timebase = oldTime;
	static int frame = 0;
	static int statframe = 0;
	static int picframe = 0;
	int newTime = glutGet(GLUT_ELAPSED_TIME);

	if (oldTime == newTime)
		return;

	frame++;
	if (newTime - timebase > 1000) {
		std::cout << "FPS: " << (int)(frame*1000.0f/(newTime-timebase)) << std::endl;
		timebase = newTime;
		frame = 0;
	}

	statframe++;
	if (statframe > 500)
	{
		printStats(statframe-1);
		statframe = 0;
	}

	picframe++;
	if (picframe == 2000)
	{
		// takes screenshot and saves to disk.
		ilutGLScreen(); 
		ilSaveImage((ILstring)("thewell-2000.png"));
	}

	int ticks;
	if (aviMode)
	{
		ticks = 1000 / aviFPS;
	} 
	else if (framebyframe) 
	{
		ticks = fbf_ticks;
	} 
	else 
	{
		if (paused || unpaused) 
		{
			ticks = fbf_ticks;
			unpaused = false;
		} 
		else 
		{
			ticks = newTime - oldTime;
		}
	}

	// update all objects
	for (std::list<IDrawable*>::iterator it = drawlist.begin(); it != drawlist.end(); it++)
	{
		(*it)->update(ticks);
	}

	oldTime = newTime;

	// render the scene
	glMatrixMode(GL_TEXTURE);
	glLoadIdentity();

	if (firstframe)
	{
		updateCubeMap(texCubeMap);
		firstframe = false;
	}
	if (moved) {
		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();
		gluLookAt(eye_coord[0], eye_coord[1], eye_coord[2],
			eye_coord[0] + line_of_sight[0], eye_coord[1] + line_of_sight[1], eye_coord[2] + line_of_sight[2],
			0.0f, 1.0f, 0.0f);
		makeInvModelViewMat();
		moved = false;
	}
	
	// display
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	// set up viewing transformation
	glViewport(0, 0, (GLsizei) vpWidth, (GLsizei) vpHeight);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(60.0, (float)vpWidth / (float)vpHeight, 1.5, 2000.0);
	
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	gluLookAt(eye_coord[0], eye_coord[1], eye_coord[2],
		eye_coord[0] + line_of_sight[0], eye_coord[1] + line_of_sight[1], eye_coord[2] + line_of_sight[2],
		0.0f, 1.0f, 0.0f);

	renderScene(false);

	glutSwapBuffers();

	if (aviMode)
	{
		static int aviframe = 0;
		unsigned char *outputImage = 0;
		outputImage = new unsigned char[vpWidth * vpHeight * 3];
	
        glReadPixels(0, 0, vpWidth, vpHeight, GL_RGB, GL_UNSIGNED_BYTE, outputImage);

		std::ostringstream oss;
		oss << "movie/file" << std::setw(4) << std::setfill('0') << aviframe << ".tga";
		writeTGA(oss.str().c_str(), outputImage);
	    delete[] outputImage;

		aviframe++;
	}
}

void Engine::renderScene(bool onlyReflecters)
{
	for (std::list<IDrawable*>::iterator it = drawlist.begin(); it != drawlist.end(); it++)
	{
		if (!onlyReflecters || (onlyReflecters && (*it)->isReflecter()))
		{
			glPushMatrix();
			(*it)->display();
			glPopMatrix();
		}
	}

	float norm_amb[] = { 0.0f, 1.0f, 0.0f, 1.0f };
	float norm_diff[] = { 0.0f, 1.0f, 0.0f, 1.0f };
	float norm_spec[] = { 1.0f, 1.0f, 1.0f, 1.0f };
	glMaterialfv(GL_FRONT, GL_DIFFUSE, norm_diff);
	glMaterialfv(GL_FRONT, GL_AMBIENT, norm_amb);
	glMaterialfv(GL_FRONT, GL_SPECULAR, norm_spec);
	glMaterialf(GL_FRONT, GL_SHININESS, 20.0);

	glPushMatrix();
	glTranslatef(120.0f, 60.0f, 80.0f);
	glutSolidCube(15.0f);
	glPopMatrix();
}

void Engine::renderSpacedBitmapString(float x, float y, int spacing, void *font, char *string)
{
	char *c;
	//int x1 = (int) x;
	for (c = string; *c != '\0'; c++) {
		glRasterPos2f(x, y);
    	glutBitmapCharacter(font, *c);
		x += glutBitmapWidth(font, *c) + (float)spacing;
	}
}

void Engine::onReshape(GLint w, GLint h)
{
	vpWidth = w;
	vpHeight = h;

	if (vpHeight == 0) vpHeight = 1;

	glViewport(0, 0, (GLsizei) vpWidth, (GLsizei) vpHeight);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(60.0, (float)vpWidth / (float)vpHeight, 1.5, 2000.0);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	
	gluLookAt(eye_coord[0], eye_coord[1], eye_coord[2],
		eye_coord[0] + line_of_sight[0], eye_coord[1] + line_of_sight[1], eye_coord[2] + line_of_sight[2],
		0.0f, 1.0f, 0.0f);
}

void Engine::onIdle()
{
	if ((!paused && !framebyframe) || (framebyframe && fbf_advance))
	{
		if (framebyframe)
		{
			fbf_advance = false;
		}
		glutPostRedisplay();
	}
}

void Engine::onKeyboard(unsigned char key, GLint x_coord, GLint y_coord)
{
	switch (key)
	{
	case 'c':
		//ilutGLScreenie(); // takes screenshot and saves to disk as a tga-file
		ilutGLScreen(); // takes screenshot and saves to disk.
		ilSaveImage((ILstring)("thewell.png")); // Skriver ikke over hvis fil med samme navn ligger der.
		break;
	case 'p':
		paused = !paused;
		if (paused)
			std::cout << "Paused..." << std::endl;
		else
		{
			std::cout << "Unpaused..." << std::endl;
			unpaused = true;
		}
		break;
	case 'f':
		framebyframe = !framebyframe;
		if (framebyframe)
		{
			std::cout << "Frame by frame mode" << std::endl;
			fbf_advance = false;
		}
		else
		{
			std::cout << "Continuous mode" << std::endl;
			unpaused = true;
		}
		break;
	case 'g':
		if (framebyframe)
		{
			fbf_advance = true;
		}
		break;

	case 'q':
	case 27:
		exit(1);

	default:
		// propagate keyboard event to all objects in key event list
		for (std::list<IKeyEvent*>::iterator it = keylist.begin(); it != keylist.end(); it++)
		{
			(*it)->onNormalKey(key);
		}
	}
}

void Engine::onSpecial(GLint key, GLint x_coord, GLint y_coord)
{
	switch (key) {
		//float angle;
		case GLUT_KEY_LEFT :
			line_of_sight[0] += sinf(0.1f);
			line_of_sight[2] -= cosf(0.1f);
			
			moved = true;
			break;
		case GLUT_KEY_RIGHT : 
			line_of_sight[0] -= sinf(0.1f);
			line_of_sight[2] += cosf(0.1f);
			moved = true;
			break;
		case GLUT_KEY_UP : 
			eye_coord[0] += 0.05f * line_of_sight[0];
			eye_coord[2] += 0.05f * line_of_sight[2];
			moved = true;
			break;
		case GLUT_KEY_DOWN : 
			eye_coord[0] -= 0.05f * line_of_sight[0];
			eye_coord[2] -= 0.05f * line_of_sight[2];
			moved = true;
			break;
	}

	// special key event
	for (std::list<IKeyEvent*>::iterator it = keylist.begin(); it != keylist.end(); it++)
	{
		(*it)->onSpecialKey(key);
	}
}

/* 
 * GLUT callbacks
 */

void gcDisplay(void)
{
	//Engine::Instance()->onDisplay();
	Engine::Instance()->render();
}

void gcReshape (GLint w, GLint h)
{
	Engine::Instance()->onReshape(w, h);
}

void gcIdle(void)
{
	Engine::Instance()->onIdle();
}

void gcKeyboard(unsigned char key, GLint x_coord, GLint y_coord)
{
	Engine::Instance()->onKeyboard(key, x_coord, y_coord);
}

void gcSpecial(GLint key, GLint x_coord, GLint y_coord)
{
	Engine::Instance()->onSpecial(key, x_coord, y_coord);
}

void Engine::makeInvModelViewMat()
{
	float mat[16];
	glGetFloatv (GL_MODELVIEW_MATRIX, mat);
	float det, oneOverDet;
	
	// compute 3*3 upper left matrix portion determinent
	det = mat[0] * ((mat[5] * mat[10]) - (mat[6] * mat[9]))
		+ mat[1] * ((mat[6] * mat[8]) - (mat[4] * mat[10]))
		+ mat[2] * ((mat[4] * mat[9]) - (mat[5] * mat[8]));
	
	oneOverDet = 1.0f / det;
	
	// compute affine part inverse
	invmat[0] = ((mat[5] * mat[10])- (mat[6] * mat[9])) * oneOverDet;
	invmat[1] = ((mat[2] * mat[9]) - (mat[1] * mat[10]))* oneOverDet;
	invmat[2] = ((mat[1] * mat[6]) - (mat[2] * mat[5])) * oneOverDet;
	invmat[3] = 0.0f;

	invmat[4] = ((mat[6] * mat[8]) - (mat[4] * mat[10]))* oneOverDet;
	invmat[5] = ((mat[0] * mat[10])- (mat[2] * mat[8])) * oneOverDet;
	invmat[6] = ((mat[2] * mat[4]) - (mat[0] * mat[6])) * oneOverDet;
	invmat[7] = 0.0f;

	invmat[8] = ((mat[4] * mat[9]) - (mat[5] * mat[8])) * oneOverDet;
	invmat[9] = ((mat[1] * mat[8]) - (mat[0] * mat[9])) * oneOverDet;
	invmat[10]= ((mat[0] * mat[5]) - (mat[1] * mat[4])) * oneOverDet;
	invmat[11]= 0.0f;

	// compute translation part inverse
	invmat[12]= -((mat[12] * invmat[0]) + (mat[13] * invmat[4]) + (mat[14] * invmat[8]));
	invmat[13]= -((mat[12] * invmat[1]) + (mat[13] * invmat[5]) + (mat[14] * invmat[9]));
	invmat[14]= -((mat[12] * invmat[2]) + (mat[13] * invmat[6]) + (mat[14] * invmat[10]));
	invmat[15]= 1.0f;
	
	invmat[12] = 0.0f;
	invmat[13] = 0.0f;
	invmat[14] = 0.0f;
	invmat[15] = 1.0f;
}

void Engine::updateCubeMap(unsigned int map)
{
	float middle = (size-1.0f)/2.0f;
	glViewport(0, 0, CUBE_MAP_SIZE, CUBE_MAP_SIZE);
	
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(90.0, 1.0, 1.0, 2000.0);

	glMatrixMode(GL_MODELVIEW);

	glPushMatrix();
	for (int i = 0; i < 6; i++) {
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
		glLoadIdentity();
		if (i == 0)			gluLookAt(middle, 0.0f, middle, middle+1.0,  0.0f, middle,      0.0f, -1.0f, 0.0f);
		else if (i == 1)	gluLookAt(middle, 0.0f, middle, middle-1.0,  0.0f, middle,      0.0f, -1.0f, 0.0f);
		else if (i == 2)	gluLookAt(middle, 0.0f, middle, middle    ,  1.0f, middle,      0.0f,  0.0f, 1.0f);
		else if (i == 3)	gluLookAt(middle, 0.0f, middle, middle    , -1.0f, middle,      0.0f,  0.0f, -1.0f);
		else if (i == 4)	gluLookAt(middle, 0.0f, middle, middle    ,  0.0f, middle+1.0f, 0.0f, -1.0f, 0.0f);
		else				gluLookAt(middle, 0.0f, middle, middle    ,  0.0f, middle-1.0f, 0.0f, -1.0f, 0.0f);

		renderScene(true);

		glEnable(GL_TEXTURE_CUBE_MAP);
		glBindTexture(GL_TEXTURE_CUBE_MAP, map);
		glCopyTexSubImage2D(GL_TEXTURE_CUBE_MAP_POSITIVE_X + i, 0,0,0,0,0, CUBE_MAP_SIZE, CUBE_MAP_SIZE);
		glDisable(GL_TEXTURE_CUBE_MAP);
	}
	glPopMatrix();
}

void Engine::printStats(int frames)
{
	ttTotal.Stop();
	unsigned long totalTime = ttTotal.TotalTime();
	std::cout << std::endl << "Statistics (" << frames << " frames)" << std::endl << std::endl;
	ttTotal.PrintTime();
	std::cout << "Frames per second: " << (frames * 1000) / totalTime << std::endl;

	for (std::list<IStatistics*>::const_iterator it = statlist.begin(); it != statlist.end(); it++)
	{
		(*it)->printStatistics();
		std::cout << std::endl;
	}
	ttTotal.Reset();
	ttTotal.Start();
}

bool Engine::writeTGA(const char *file, unsigned char *outImage)
{
	FILE *pFile = 0;
	unsigned char tgaHeader[12] = {0, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0};
	unsigned char header[6];
	unsigned char bits = 0;
	int colorMode = 0;
	unsigned char tempColors = 0;

	// Open file for output.
	pFile = fopen(file, "wb");

	// Check if the file opened or not.
	if(!pFile) { fclose(pFile); return false; }

	// Set the color mode, and the bit depth.
	colorMode = 3;
	bits = 24;

	// Save the width and height.
	header[0] = vpWidth % 256;
	header[1] = vpWidth / 256;
	header[2] = vpHeight % 256;
	header[3] = vpHeight / 256;
	header[4] = bits;
	header[5] = 0;

	// Write the headers to the top of the file.
	fwrite(tgaHeader, sizeof(tgaHeader), 1, pFile);
	fwrite(header, sizeof(header), 1, pFile);

	// Now switch image from RGB to BGR.
	for (int index = 0; index < vpWidth * vpHeight * colorMode; index += colorMode)
	{
		tempColors = outImage[index];
		outImage[index] = outImage[index + 2];
		outImage[index + 2] = tempColors;
	}

	// Finally write the image.
	fwrite(outImage, vpHeight * vpWidth * colorMode, 1, pFile);

	// close the file.
	fclose(pFile);
	return true;
}
