#include "StdAfx.h"
#include "FBO.h"

FBO::FBO(unsigned int fboWidth, unsigned int fboHeight)
: width(fboWidth), height(fboHeight), fboId(0), bound(false)
{
	glGenFramebuffersEXT(1, &fboId);
	rgbaData = new float[width*height*4];
}

FBO::~FBO()
{
	if (bound)
		glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
	glDeleteFramebuffersEXT(1, &fboId);
	delete[] rgbaData;
}

void FBO::bind()
{
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fboId);
	bound = true;
}

void FBO::unbind()
{
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
	bound = false;
}

/**
 * Checks framebuffer status.
 * Copied directly out of the spec, modified to deliver a return value.
 */
bool FBO::checkFBStatus() {
    GLenum status = (GLenum) glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT);

    switch (status) {
        case GL_FRAMEBUFFER_COMPLETE_EXT:
            return true;
        case GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT_EXT:
			std::cerr << "Framebuffer incomplete, incomplete attachment" << std::endl;
            return false;
        case GL_FRAMEBUFFER_UNSUPPORTED_EXT:
			std::cerr << "Unsupported framebuffer format" << std::endl;
            return false;
        case GL_FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT_EXT:
			std::cerr << "Framebuffer incomplete, missing attachment" << std::endl;
            return false;
        case GL_FRAMEBUFFER_INCOMPLETE_DIMENSIONS_EXT:
			std::cerr << "Framebuffer incomplete, attached images must have same dimensions" << std::endl;
            return false;
        case GL_FRAMEBUFFER_INCOMPLETE_FORMATS_EXT:
			std::cerr << "Framebuffer incomplete, attached images must have same format" << std::endl;
            return false;
        case GL_FRAMEBUFFER_INCOMPLETE_DRAW_BUFFER_EXT:
			std::cerr << "Framebuffer incomplete, missing draw buffer" << std::endl;
            return false;
        case GL_FRAMEBUFFER_INCOMPLETE_READ_BUFFER_EXT:
			std::cerr << "Framebuffer incomplete, missing read buffer" << std::endl;
            return false;
		default:
			std::cerr << "Unknown frame buffer error status" << std::endl;
			return false;
    }
}

bool FBO::attach(unsigned int attachment, unsigned int textureId)
{
	if (!bound)
	{
#ifdef _DEBUG
		std::cerr << "FBO not bound" << std::endl;
#endif
		return false;
	}

	glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT + attachment, 
		GL_TEXTURE_2D, textureId, 0);
	return checkFBStatus();
}

void FBO::unattach(unsigned int attachment)
{
	if (!bound)
	{
#ifdef _DEBUG
		std::cerr << "FBO not bound" << std::endl;
#endif
		return;
	}
	
	glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT + attachment, 
		GL_TEXTURE_2D, 0, 0);
}

void FBO::read(unsigned int attachment, float *data, int offset)
{
	if (!bound)
	{
#ifdef _DEBUG
		std::cerr << "FBO not bound" << std::endl;
#endif
		return;
	}

	//float *rgbaData = new float[width*height*4];
	glReadBuffer(GL_COLOR_ATTACHMENT0_EXT + attachment);
	glReadPixels(0, 0, width, height, GL_RGBA, GL_FLOAT, rgbaData);
	
	for (unsigned int i = 0; i < width*height; i++)
	{
		data[i] = rgbaData[i*4+offset];
	}
	//delete[] rgbaData;
}

void FBO::read(unsigned int attachment, float *rdata, float *gdata, float *bdata, float *adata)
{
	if (!bound)
	{
#ifdef _DEBUG
		std::cerr << "FBO not bound" << std::endl;
#endif
		return;
	}

	//float *rgbaData = new float[width*height*4];
	glReadBuffer(GL_COLOR_ATTACHMENT0_EXT + attachment);
	glReadPixels(0, 0, width, height, GL_RGBA, GL_FLOAT, rgbaData);
	
	for (unsigned int i = 0; i < width*height; i++)
	{
		if (rdata != NULL) rdata[i] = rgbaData[i*4];
		if (gdata != NULL) gdata[i] = rgbaData[i*4+1];
		if (bdata != NULL) bdata[i] = rgbaData[i*4+2];
		if (adata != NULL) adata[i] = rgbaData[i*4+3];
	}
	//delete[] rgbaData;
}

void FBO::readPacked(unsigned int attachment, float *data)
{
	if (!bound)
	{
#ifdef _DEBUG
		std::cerr << "FBO not bound" << std::endl;
#endif
		return;
	}

	//float *tdata = new float[width*height*4];
	glReadBuffer(GL_COLOR_ATTACHMENT0_EXT + attachment);
	glReadPixels(0, 0, width, height, GL_RGBA, GL_FLOAT, rgbaData);
	
	// map packed texture rgba values to data array
/*	for (unsigned int i = 0; i < width; i++)
	{
		for (unsigned int j = 0; j < height; j++)
		{
			int tpos = (i*width + j)*4;
			int apos = i*width*4 + j*2;

			data[apos] = tdata[tpos];
			data[apos+1] = tdata[tpos+1];
			data[apos+width*2] = tdata[tpos+2];
			data[apos+width*2+1] = tdata[tpos+3];
		}
	}*/
	for (unsigned int i = 0; i < width*height*4; i++)
	{
		data[i] = rgbaData[i];
	}
	//delete[] tdata;
}