#include "StdAfx.h"

#include <sstream>
#include <string>
#include "Shader.h"

using namespace std;

ShaderProgram::ShaderProgram()
	: _program(0)
	, _vertexShader(0)
	, _fragmentShader(0)
{}

void ShaderProgram::addVertexSource(const char *filename)
{
	string source = readSource(filename);
	if (source.length() == 0)
		return;

	_vertexSources.push_back(source);
}

void ShaderProgram::addFragmentSource(const char *filename)
{
	string source = readSource(filename);
	if (source.length() == 0)
		return;

	_fragmentSources.push_back(source);
}

void ShaderProgram::load()
{
	if (_program == 0)
	{
		if (_vertexShader == 0 && _vertexSources.size() > 0)
		{
			if (!compileVertexShader())
				return;
		}
		if (_fragmentShader == 0 && _fragmentSources.size() > 0)
		{
			if (!compileFragmentShader())
				return;
		}
		if (_vertexShader == 0 && _fragmentShader == 0)
			return;

		if (!link())
			return;
	}

	// we have a linked program object if we get this far
	glUseProgram(_program);
}

void ShaderProgram::unload()
{
	// go back to fixed functionality
	glUseProgram(0);
}

bool ShaderProgram::compileVertexShader()
{
	int nstrings = (int) _vertexSources.size();
	const GLchar **sources = new const GLchar*[nstrings];
	int i = 0;
	for (list<string>::const_iterator it = _vertexSources.begin(); it != _vertexSources.end(); it++)
	{
		sources[i++] = it->c_str();
	}

	// create shader
	_vertexShader = glCreateShader(GL_VERTEX_SHADER);
	glShaderSource(_vertexShader, nstrings, sources, NULL);
	
	// delete sources
	delete[] sources;

	bool result = compileShader(_vertexShader);
	if (!result)
	{
		// clean up
		glDeleteShader(_vertexShader);
		_vertexShader = 0;
	}
	return result;
}

bool ShaderProgram::compileFragmentShader()
{
	int nstrings = (int) _fragmentSources.size();
	const GLchar **sources = new const GLchar*[nstrings];
	int i = 0;
	for (list<string>::const_iterator it = _fragmentSources.begin(); it != _fragmentSources.end(); it++)
	{
		sources[i++] = it->c_str();
	}

	// create shader
	_fragmentShader = glCreateShader(GL_FRAGMENT_SHADER);
	glShaderSource(_fragmentShader, nstrings, sources, NULL);
	
	// delete sources
	delete[] sources;

	bool result = compileShader(_fragmentShader);
	if (!result)
	{
		// clean up
		glDeleteShader(_fragmentShader);
		_fragmentShader = 0;
	}
	return result;
}

bool ShaderProgram::compileShader(GLuint shader)
{
	// this function requires the shader source to be loaded into the shader object first

	// compile shader
	glCompileShader(shader);

	// check compile status
	int compileStatus;
	glGetShaderiv(shader, GL_COMPILE_STATUS, &compileStatus);
	if (compileStatus == GL_FALSE)
	{
		int length;
		glGetShaderiv(shader, GL_INFO_LOG_LENGTH, &length);
		if (length > 0)
		{
			GLchar *infoLog = new GLchar[length];
			glGetShaderInfoLog(shader, length, NULL, infoLog);
			std::cerr << "Compilation of shader failed: " << infoLog << std::endl;
		}
	}
	return (compileStatus == GL_TRUE);
}

bool ShaderProgram::link()
{
	// this function assumes the shader objects has been compiled
	if (!_vertexShader && !_fragmentShader)
		// no shaders to attach
		return false;

	_program = glCreateProgram();
	if (_vertexShader)
	{
		glAttachShader(_program, _vertexShader);
		glDeleteShader(_vertexShader); // marked for deletion until detached from program
	}
	if (_fragmentShader)
	{
		glAttachShader(_program, _fragmentShader);
		glDeleteShader(_fragmentShader); // marked for deletion until detached from program
	}

	glLinkProgram(_program);

	// check link status
	int linkStatus;
	glGetProgramiv(_program, GL_LINK_STATUS, &linkStatus);
	if (linkStatus == GL_FALSE)
	{
		int length;
		glGetProgramiv(_program, GL_INFO_LOG_LENGTH, &length);
		if (length > 0)
		{
			GLchar *infoLog = new GLchar[length];
			glGetProgramInfoLog(_program, length, NULL, infoLog);
			std::cerr << "Linking of program failed: " << infoLog << std::endl;

			// clean up
			glDeleteProgram(_program);
			_program = 0;
		}
	}
	return (linkStatus == GL_TRUE);
}

string ShaderProgram::readSource(const char *filename)
{
	ifstream file;
	ostringstream source;
	
	file.open(filename);
	if (!file)
	{
		return "";
	}

	string line;
	getline(file, line);
	while (file)
	{
		source << line << endl;
		getline(file, line);
	}
	file.close();

	return source.str();
}

GPShaderProgram::GPShaderProgram()
	: ShaderProgram()
	, _width(0)
	, _height(0)
{
	memset(_originalView, 0, sizeof(int)*4);
}

GPShaderProgram::GPShaderProgram(int width, int height)
	: ShaderProgram()
	, _width(width)
	, _height(height)
{
	memset(_originalView, 0, sizeof(int)*4);
}

void GPShaderProgram::initialize2D(int width, int height)
{
	_width = width;
	_height = height;

	// setup viewport for 1:1 pixel:texel
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluOrtho2D(0.0, (float) width, 0.0, (float) height);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glViewport(0, 0, width, height);

	glPolygonMode(GL_FRONT, GL_FILL);
    glDisable(GL_CULL_FACE);
	glDisable(GL_DEPTH_TEST);
	glDisable(GL_LIGHTING);
}

void GPShaderProgram::reset()
{
	glActiveTexture(GL_TEXTURE0);
	glViewport(_originalView[0], _originalView[1], _originalView[2], _originalView
		[3]);
    glEnable(GL_CULL_FACE);
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_LIGHTING);
}