#pragma once

class ShaderProgram
{
public:
	ShaderProgram();

	void addVertexSource(const char *filename);
	void addFragmentSource(const char *filename);
	void load();
	void unload();

	/* --- uniform variabel setters --- */
	GLint getUniformLocation(const char *name);
	void setUniform1i(const char *name, GLint i);
	void setUniform1f(const char *name, GLfloat f);
	void setUniform3fv(const char *name, GLuint count, const GLfloat *f);
	void setUniform3f(const char *name, GLfloat f1, GLfloat f2, GLfloat f3);
	void setSampler2D(const char *name, GLenum textureUnit, GLuint texture);

private:
	std::string readSource(const char *filename);
	bool compileVertexShader();
	bool compileFragmentShader();
	bool compileShader(GLuint shader);
	bool link();

	std::list<std::string> _vertexSources;
	std::list<std::string> _fragmentSources;

	GLuint _program;
	GLuint _vertexShader;
	GLuint _fragmentShader;
};

class GPShaderProgram : public ShaderProgram
{
public:
	GPShaderProgram();
	GPShaderProgram(int width, int height);

	void initialize2D(int width, int height);
	void draw2D(unsigned int attachment);
	void reset();

private:
	int _width;
	int _height;
	int _originalView[4];
};

/*--- inline functions ---*/

inline
GLint ShaderProgram::getUniformLocation(const char *name)
{
#ifndef _DEBUG
	return glGetUniformLocation(_program, name);
#else
	if (_program == 0)
	{
		std::cerr << "Program must be linked first." << std::endl;
		return -1; // has to be linked first
	}

	GLint loc = glGetUniformLocation(_program, name);
	if (loc == -1)
	{
		std::cerr << "No such uniform named " << name << std::endl;
	}
	return loc;
#endif
}

inline
void ShaderProgram::setUniform1i(const char *name, GLint i)
{
	glUniform1i(getUniformLocation(name), i);
}

inline
void ShaderProgram::setUniform1f(const char *name, GLfloat f)
{
	glUniform1f(getUniformLocation(name), f);
}

inline
void ShaderProgram::setUniform3fv(const char *name, GLuint count, const GLfloat *f)
{
	glUniform3fv(getUniformLocation(name), count, f);
}

inline
void ShaderProgram::setUniform3f(const char *name, GLfloat f1, GLfloat f2, GLfloat f3)
{
	glUniform3f(getUniformLocation(name), f1, f2, f3);
}

inline
void ShaderProgram::setSampler2D(const char *name, GLenum textureUnit, GLuint texture)
{
	glActiveTexture(textureUnit);
	glBindTexture(GL_TEXTURE_2D, texture);
	setUniform1i(name, textureUnit - GL_TEXTURE0);
}

inline
void GPShaderProgram::draw2D(unsigned int attachment)
{
	glDrawBuffer(GL_COLOR_ATTACHMENT0_EXT + attachment);

	glBegin(GL_QUADS);
		glTexCoord2f(0.0f, 0.0f); glVertex2f(0.0f, 0.0f);
		glTexCoord2f(1.0f, 0.0f); glVertex2f((float) _width, 0.0f);
		glTexCoord2f(1.0f, 1.0f); glVertex2f((float) _width, (float) _height);
		glTexCoord2f(0.0f, 1.0f); glVertex2f(0.0f, (float) _height);
	glEnd();
}
