#include "StdAfx.h"
#include "Engine.h"
#include "WaterSurface.h"

WaterSurface::WaterSurface(unsigned int surfaceSize, float spacing, float waterLevel)
: drawmode(GL_FILL), drawHeightMap(false), drawSpeedU(false), drawSpeedV(false),
ttDeparture("Compute departure"), ttPrepSolver("Prepare solver"), ttSolver("Solver"), ttPostSolver("Post solver"),
ttDraw("Draw water surface")
{
	size = surfaceSize;
	size_2 = size*size;
	dS = spacing;
	dS_1 = 1.0f / dS;
	dS_2 = dS*dS;
	dS_2_1 = 1.0f / dS_2;
	g = 9.81f;

	position[0] = 0.0f;
	position[1] = 0.0f;
	position[2] = 0.0f;

	reflectionShader.addFragmentSource("reflectionRefractionShader.frag");
	reflectionShader.addVertexSource("reflectionRefractionShader.vert");

	A = new float[size_2];
	b = new float[size_2];
	h = new float[size_2];
	u = new float[size_2];
	v = new float[size_2];
	x = new float[size];
	z = new float[size];
	tildeH = new float[size_2];
	tildeU = new float[size_2];
	tildeV = new float[size_2];
	normals = new float[size_2*3];
	noise = new float[size_2];

	for (unsigned int i = 0; i < size; i++)
	{
		x[i] = i * dS;
		z[i] = i * dS;
	}

	for(int i = 0; i < 512 ; i++) perm[i] = p[i & 255];
	

	for (unsigned int i = 0; i < size_2; i++)
	{
		A[i] = /*A[i*5+1] = A[i*5+2] = A[i*5+3] = A[i*5+4] = */0.0f;
		h[i] = waterLevel;
		u[i] = 0.0f;
		v[i] = 0.0f;
	}

	for (unsigned int i = 0; i < size; i++)
		for (unsigned int j = 0; j < size; j++)
			noise[i*size + j] = simplexNoise(i, j)*0.1;
}

WaterSurface::~WaterSurface()
{
	delete[] A;
	delete[] b;
	delete[] h;
	delete[] u;
	delete[] v;
	delete[] x;
	delete[] z;
	delete[] tildeH;
	delete[] tildeU;
	delete[] tildeV;
}

void WaterSurface::setHeight(unsigned int x, unsigned int z, float value)
{
	if (x < size && z < size)
		h[x*size+z] = value > 0.0f ? value : 0.0f;
}

void WaterSurface::setSpeed(unsigned int x, unsigned int z, float uval, float vval)
{
	u[x*size+z] = uval;
	v[x*size+z] = vval;
}

void WaterSurface::checkGLErrors(const char *label) 
{
#ifdef _DEBUG
    GLenum errCode;
    const GLubyte *errStr;
    
    if ((errCode = glGetError()) != GL_NO_ERROR) {
		errStr = gluErrorString(errCode);
		std::cerr << "OpenGL ERROR: " << errStr << " (Label: " << label << ")" << std::endl;
    }
#endif
}

void WaterSurface::update(int ticks)
{
	//dT = ticks / 1000.0f; // 0.001f; // 
	dT = 1.0f / 50.0f; //0.012f; // TODO for  ha en fast verdi  se p

	checkGLErrors("pre departure");
	ttDeparture.Start();
	computeDeparture();
	ttDeparture.Stop();
	checkGLErrors("departure");

	ttPrepSolver.Start();
	prepareSolver();
	ttPrepSolver.Stop();
	checkGLErrors("prepare solver");

	ttSolver.Start();
	solver();
	ttSolver.Stop();
	checkGLErrors("solver");

	ttPostSolver.Start();
	postSolver();
	ttPostSolver.Stop();
	checkGLErrors("post solver");
}

void WaterSurface::computeDeparture()
{
	for (unsigned int i = 0 ; i < size; i++) {
		for (unsigned int j = 0; j < size; j++) {
			int pos = i*size + j;

			// departure point from (31)
			float aX = dT * u[pos];
			float aZ = dT * v[pos];

			float depX = (x[i]-aX) * dS_1;
			float depZ = (z[j]-aZ) * dS_1;

			int m_1, m, n_1, n; // table positions
			m = (int)(depX*dS + 1);
			m_1 = m - 1;
			n = (int)(depZ*dS + 1);
			n_1 = n - 1;

			if (m_1 < 0)
			{
				m_1 = 0;
				m = 1;
			}
			else if (m >= (int) size)
			{
				m_1 = size - 2;
				m = size - 1;
			}

			if (n_1 < 0)
			{
				n_1 = 0;
				n = 1;
			}
			else if (n >= (int) size)
			{
				n_1 = size - 2;
				n = size - 1;
			}
			
			int posMN = m*size + n;
			int posM1N = m_1*size + n;
			int posMN1 = m*size + n_1;
			int posM1N1 = m_1*size + n_1;

			if (m == m_1) {
				if (n == n_1) {
					tildeH[pos] = h[posMN];
					tildeU[pos] = u[posMN];
					tildeV[pos] = v[posMN];
				} else {
					float z1 = z[n]*dS_1 - depZ; // (z[n] - z[j] + aZ) * dS_1;
					float z2 = 1.0f - z1; //(z[j]-aZ-z[n_1])*dS_1;
					tildeH[pos] = z1 * h[posMN1] + z2 * h[posMN];
					tildeU[pos] = z1 * u[posMN1] + z2 * u[posMN];
					tildeV[pos] = z1 * v[posMN1] + z2 * v[posMN];
				}
			} else {
				float x1 = x[m]*dS_1 - depX; //(x[m] - x[i] + aX) * dS_1;
				float x2 = 1.0f - x1; //(x[i] - aX - x[xm_1]) * dS_1;
				if (n == n_1) {
					tildeH[pos] = x1 * h[posM1N] + x2 * h[posMN];
					tildeU[pos] = x1 * u[posM1N] + x2 * u[posMN];
					tildeV[pos] = x1 * v[posM1N] + x2 * v[posMN];
				} else {
					float z1 = z[n]*dS_1 - depZ; //(z[n] - z[j] + aZ) * dS_1;
					float z2 = 1.0f - z1; //(z[j]-aZ-z[zn_1])*dS_1;
					tildeH[pos] = x2 * (z1 * h[posMN1] + z2 * h[posMN]) + x1 * (z1 * h[posM1N1] + z2 * h[posM1N]);
					tildeU[pos] = x2 * (z1 * u[posMN1] + z2 * u[posMN]) + x1 * (z1 * u[posM1N1] + z2 * u[posM1N]);
					tildeV[pos] = x2 * (z1 * v[posMN1] + z2 * v[posMN]) + x1 * (z1 * v[posM1N1] + z2 * v[posM1N]);
				}
			}
		}
	}
}

void WaterSurface::prepareSolver()
{
	float t2g = dT*dT*g;

	/* A simplification of (25). Assumes flat bottom */
	float constant = dS_2 / (t2g);

	for (unsigned int i = 0; i < size; i++)
	{
		for (unsigned int j = 0; j < size; j++)
		{
			int pos = i*size + j;

			A[pos] = 4.0f + constant / h[pos];
			b[pos] = (tildeH[pos] * constant / h[pos])
				- ((dS / (2*dT*g)) * (tildeU[pos+(i<size-1?size:0)] - tildeU[pos-(i>0?size:0)] + tildeV[pos+(j<size-1?1:0)] - tildeV[pos-(j>0?1:0)]));
		}
	}
}

void WaterSurface::postSolver()
{
	// updates h, u and v.

	for (unsigned int i = 1; i < size-1; i++) {
		h[i] = 2.0f * h[size + i] - h[2*size + i];
		h[(size-1)*size + i] = 2.0f * h[(size-2)*size + i] - h[(size-3)*size + i];
		h[i*size] = 2.0f * h[i*size + 1] - h[i*size + 2];
		h[i*size + size-1] = 2.0f * h[i*size + size-2] - h[i*size + size-3];
	}
	
	h[0] = 2.0f * h[size + 1] - h[2*size + 2];
	h[size-1] = 2.0f * h[size + size-2] - h[2*size + size-3];
	h[(size-1)*size] = 2.0f * h[(size-2)*size + 1] - h[(size-3)*size + 2];
	h[(size-1)*size + size-1] = 2.0f * h[(size-2)*size + size-2] - h[(size-3)*size + size-3];

	for (unsigned int i = 0; i < size; i++) {
		for (unsigned int j = 0; j < size; j++) {
			if (h[i*size + j] < 0.0f)
				h[i*size + j] = 0.0f;
		}
	}

	float halfdTg = -0.5f * dT * g;
	for (unsigned int i = 0; i < size; i++) {
		for (unsigned int j = 0; j < size; j++) {
			int pos = i*size + j;
			if (i == 0 || i == size-1 || j == 0 || j == size-1) {
				u[pos] = 0.0f;
				v[pos] = 0.0f;
			} else {
				u[pos] = halfdTg*(h[pos+size] - h[pos-size])*dS_1 + tildeU[pos];
				v[pos] = halfdTg*(h[pos+1] - h[pos-1])*dS_1 + tildeV[pos];
			}
		}
	}
}

void WaterSurface::makeNormals()
{
	float normS = 2.0f * dS;

 	for (unsigned int i = 1; i < size-1; i++) {
		for (unsigned int j = 1; j < size-1; j++) {

			unsigned int pos = i*size + j;
			unsigned int pos3 = pos*3;

			normals[pos3] = (h[pos-size] - h[pos+size]) + noise[pos];
			normals[pos3+1] = normS;
			normals[pos3+2] = (h[pos-1] - h[pos+1]) + noise[pos];
			
			float length = sqrtf(normals[pos3]*normals[pos3] + normals[pos3+1]*normals[pos3+1] 
				+ normals[pos3+2]*normals[pos3+2]);
			normals[pos3] /= length;
			normals[pos3+1] /= length;
			normals[pos3+2] /= length;
		}
	}

	// setup normals for edges too
	for (unsigned int i = 0; i < size; i++) {
		unsigned int pos = i*size*3;
		normals[pos] = normals[pos+3];
		normals[pos+1] = normals[pos+4];
		normals[pos+2] = normals[pos+5];

		pos = (i*size + size-1)*3;
		normals[pos] = normals[pos-3];
		normals[pos+1] = normals[pos-2];
		normals[pos+2] = normals[pos-1];
	}
	for (unsigned int i = 0; i < size; i++) {
		unsigned int toPos = i*3;
		unsigned int fromPos = i*3 + size*3;
		normals[toPos] = normals[fromPos];
		normals[toPos+1] = normals[fromPos+1];
		normals[toPos+2] = normals[fromPos+2];

		toPos = (size-1 + i)*3;
		fromPos = (size-2 + i)*3;
		normals[toPos] = normals[fromPos];
		normals[toPos+1] = normals[fromPos+1];
		normals[toPos+2] = normals[fromPos+2];
	}
}

void WaterSurface::drawNormals()
{
	glPushMatrix();
	glTranslatef(position[0], position[1], position[2]);
	float norm_amb[] = { 0.0f, 0.0f, 0.0f, 1.0f };
	float norm_diff[] = { 0.0f, 0.0f, 0.0f, 1.0f };
	float norm_spec[] = { 0.0f, 0.0f, 0.0f, 1.0f };
	glMaterialfv(GL_FRONT, GL_DIFFUSE, norm_diff);
	glMaterialfv(GL_FRONT, GL_AMBIENT, norm_amb);
	glMaterialfv(GL_FRONT, GL_SPECULAR, norm_spec);
	glMaterialf(GL_FRONT, GL_SHININESS, 0.0);

	int pos = 0;
	for (unsigned int i = 0; i < size; i += 4) {
		for (unsigned int j = 0; j < size; j += 4) {
			pos = i*size + j;
			glBegin(GL_LINES);
				glVertex3f(x[i], h[pos], z[j]);
				glVertex3f(x[i]+3.0f*normals[pos*3], h[pos]+3.0f*normals[pos*3+1], z[j]+3.0f*normals[pos*3+2]);
			glEnd();
		}
	}
	glPopMatrix();
}

void WaterSurface::display()
{
	ttDraw.Start();
	makeNormals();
	//drawNormals();

	glMatrixMode(GL_TEXTURE);
	glLoadMatrixf(Engine::Instance()->GetInvModelViewMatrix());
	glMatrixMode(GL_MODELVIEW);

	glEnable(GL_TEXTURE_CUBE_MAP);
	glActiveTexture(GL_TEXTURE0);
	glBindTexture(GL_TEXTURE_CUBE_MAP, Engine::Instance()->GetTexCubeMap());

	reflectionShader.load();
	reflectionShader.setUniform1i("CubeMap", 0);
	reflectionShader.setUniform1f("WaterSize", size);
	reflectionShader.setUniform3fv("EyeCoord", 1, Engine::Instance()->GetEyeCoord());

	drawWater();

	reflectionShader.unload();

	glDisable(GL_TEXTURE_CUBE_MAP);
	int x = 0;
	if (drawHeightMap)
	{
		overlayHeightMap(x, 0);
		x += (size+2)*3;
	}
	if (drawSpeedU)
	{
		overlaySpeedU(x, 0);
		x += (size+2)*3;
	}
	if (drawSpeedV)
	{
		overlaySpeedV(x, 0);
		x += (size+2)*3;
	}
	ttDraw.Stop();
}

void WaterSurface::drawWater()
{
	glPushMatrix();
	//glTranslatef(64.0, 50, 64.0);

	glPolygonMode(GL_FRONT, drawmode);
	
	float spacing = 1.0f / (size * dS);
	//int pos = size;
	int pos = 0;

	for (unsigned int i = 0; i < size-1; i++) {
		glBegin(GL_TRIANGLE_STRIP);

		for (unsigned int j = 0; j < size; j++) {
			glTexCoord2f(((float)(i+1))/(float)size, (float)(j)/(float)size);
			glNormal3fv(normals + (pos+size)*3);
			glVertex3f(x[i+1], h[pos+size], z[j]);
			
			glTexCoord2f((float)(i)/(float)size, (float)(j)/(float)size);
			glNormal3fv(normals + pos*3);
			glVertex3f(x[i], h[pos], z[j]);

			pos++;
		}
		glEnd();	
	}

	glPopMatrix();
}

WaterSurfaceJacobi::WaterSurfaceJacobi(unsigned int surfaceSize, float spacing, float waterLevel)
: WaterSurface(surfaceSize, spacing, waterLevel)
{
}

void WaterSurfaceJacobi::solver()
{
	jacobi(5, dT/100000.0f);
}

void WaterSurfaceJacobi::jacobi(int imax, float tolerance)
{
	float *new_h = new float[size_2];
	int iter_num = 0;
	float sqrtolr = tolerance*tolerance;
	float sqrdiff;

	do {
		for (unsigned int i = 0; i < size; i++) {
			for (unsigned int j = 0; j < size; j++) {
				int pos = i*size + j;

				new_h[pos] = (b[pos] + 4.0f*h[pos]) / A[pos];
			}
		}

		// check for convergence
		sqrdiff = sqrdist(new_h, h);

		// copy the new height map to the height map
		memcpy(h, new_h, sizeof(float) * size_2);

		iter_num++;
	} while ((iter_num < imax) && (sqrdiff >= sqrtolr));
	delete[] new_h;
}

float WaterSurfaceJacobi::sqrdist(float *x, float *y) 
{
	float sum = 0.0;

	for (unsigned int i = 0; i < size_2; i++) {
		float value = x[i] - y[i];
		sum += value * value;
	}
	return sum;
}

WaterSurfaceLU::WaterSurfaceLU(unsigned int surfaceSize, float spacing, float waterLevel)
: WaterSurface(surfaceSize, spacing, waterLevel)
{
	a_mat = new float[size_2*size_2];
	lower = new float[size_2*size_2];
	upper = new float[size_2*size_2];
}

WaterSurfaceLU::~WaterSurfaceLU()
{
	delete []a_mat;
	delete []lower;
	delete []upper;
}

void WaterSurfaceLU::solver()
{
	a_to_matrix();
	LU_decomp();
	LU_solve();
}

void WaterSurfaceLU::a_to_matrix()
{
	for (unsigned int i = 0; i != size_2; i++) {
		for (unsigned int j = 0; j != size_2; j++) a_mat[i*size_2 + j] = 0.0f;
		int pos = i*size_2 + i;
		a_mat[pos] = A[i*5];
		if (i < size_2-1)	 a_mat[pos + 1]		= A[i*5 + 1];
		if (i < size_2-size) a_mat[pos + size]	= A[i*5 + 2];
		if (i >= size)		 a_mat[pos - size]	= A[i*5 + 3];
		if (i >= 1)			 a_mat[pos - 1]		= A[i*5 + 4];
	}
}

void WaterSurfaceLU::LU_decomp()
{
	for (unsigned int i = 0; i < size_2; i++) {
		for (unsigned int j = 0; j < size_2; j++) {
			if (i == j) lower[i*size_2 + j] = 1.0f;
			else		lower[i*size_2 + j] = 0.0f;
			upper[i*size_2 + j] = 0.0f;
		}
	}

	for (unsigned int k = 0; k < size_2; k++) {
		upper[k*size_2 + k] = a_mat[k*size_2 + k];
		for (unsigned int i = k; i < size_2; i++) {
			lower[i*size_2 + k] = a_mat[i*size_2 + k] / upper[k*size_2 + k];
			upper[k*size_2 + i] = a_mat[k*size_2 + i];
		}
		for (unsigned int i = k; i < size_2; i++) {
			for (unsigned int j = k; j < size_2 ; j++)
				a_mat[i*size_2 + j] -= lower[i*size_2 + k] * upper[k*size_2 + j];
		}
	}
}

void WaterSurfaceLU::LU_solve()
{
	float *local_y = new float[size_2];
	float *local_x = new float[size_2];

	for (unsigned int i = 0; i < size_2; i++) {
		local_y[i] = b[i];
		for (unsigned int j = 0; j < i; j++)
			local_y[i] -= lower[i*size_2 + j] * local_y[j];
	}
	for (int i = size_2 - 1; i >= 0; i--) {
		local_x[i] = local_y[i];
		for (unsigned int j = i + 1; j < size_2; j++)
			local_x[i] -= upper[i*size_2 + j] * local_x[j];
		local_x[i] /= upper[i*size_2 + i];
	}

	for (unsigned int i = 0; i < size_2; i++)
		h[i] = local_x[i];
}

void WaterSurface::onNormalKey(unsigned char key)
{
	switch (key)
	{
	case 'm':
		switch (drawmode)
		{
		case GL_FILL:
			drawmode = GL_LINE;
			break;
		case GL_POINT:
			drawmode = GL_FILL;
			break;
		case GL_LINE:
		default:
			drawmode = GL_POINT;
		}
		glutPostRedisplay();
		break;
	case 'h':
		drawHeightMap = !drawHeightMap;
		glutPostRedisplay();
		break;
	case 'u':
		drawSpeedU = !drawSpeedU;
		glutPostRedisplay();
		break;
	case 'v':
		drawSpeedV = !drawSpeedV;
		glutPostRedisplay();
		break;
	}
}

void WaterSurface::onSpecialKey(int key)
{
}

void WaterSurface::printHeight()
{
	for (unsigned int i = 0; i < size; i++)
	{
		std::cout << "\n[" << i << "] ";
		for (unsigned int j = 0; j < size; j++)
		{
			std::cout << h[i*size+j] << " ";
		}
	}
}

void WaterSurface::printTable(const char *name, const float *a)
{
	std::cout << name << std::endl;
	for (unsigned int i = 0; i < size; i++)
	{
		std::cout << "[" << i << "] ";
		for (unsigned int j = 0; j < size; j++)
		{
			std::cout << a[i*size + j] << " ";
		}
		std::cout << std::endl;
	}
}

void WaterSurface::overlayHeightMap(int x, int y)
{
	overlayTable2D(x, y, h, size, 0.0f, 0.0f, 1.0f);
}

void WaterSurface::overlaySpeedU(int x, int y)
{
	overlayTable2D(x, y, u, size, 1.0f, 0.0f, 0.0f);
}

void WaterSurface::overlaySpeedV(int x, int y)
{
	overlayTable2D(x, y, v, size, 0.0f, 1.0f, 0.0f);
}

void WaterSurface::overlayTable2D(int x, int y, float *table, unsigned int tableSize, float r, float g, float b)
{
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();

	glViewport(x, y, (tableSize+2)*3, (tableSize+2)*3);
	gluOrtho2D(0.0, (double) tableSize+2, (double) tableSize+2, 0.0);
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();
	glDisable(GL_CULL_FACE);
	glDisable(GL_DEPTH_TEST);
	glDisable(GL_LIGHTING);
	glPointSize(3.0f);
	glLineWidth(1.0f);

	// determine max and min heights
	float maxVal = 0.0f;
	float minVal = 100000.0f;
	for (unsigned int i = 0; i < tableSize*tableSize; i++)
	{
		if (table[i] > maxVal) maxVal = table[i];
		if (table[i] < minVal) minVal = table[i];
	}
	float delta = maxVal - minVal;
	if (delta < 1.0f) delta = 1.0f;

	glBegin(GL_LINE_LOOP);
		glColor3f(0.0f, 0.0f, 0.0f);
		glVertex2f(0.0f, 0.0f);
		glVertex2f((float) tableSize+1, 0.0f);
		glVertex2f((float) tableSize+1, (float) tableSize+1);
		glVertex2f(0.0f, (float) tableSize+1);
	glEnd();
	glBegin(GL_POINTS);
		for (unsigned int i = 0; i < tableSize; i++)
		{
			for (unsigned int j = 0; j < tableSize; j++)
			{
				float value = (table[i*tableSize+j] - minVal) / delta;
				glColor3f(r*value, g*value, b*value); // (h[i*size+j] - 5.0f) / 5.0f); 
				glVertex2f((float) i+1, (float) j+1);
			}
		}
	glEnd();

	glEnable(GL_CULL_FACE);
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_LIGHTING);
	glPopMatrix();
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
}

float WaterSurface::simplexNoise(float xin, float yin) 
{
	float n0, n1, n2; // Noise contributions from the three corners

	// Skew the input space to determine which simplex cell we're in
	const float F2 = 0.5*(sqrt(3.0)-1.0);
	float s = (xin+yin)*F2; // Hairy factor for 2D
	int i = (xin+s) > 0 ? (int)(xin+s) : (xin+s) - 1;
	int j = (yin+s) > 0 ? (int)(yin+s) : (yin+s) - 1;

	const float G2 = (3.0-sqrt(3.0))/6.0;
	float t = (i+j)*G2;
	float X0 = i-t; // Unskew the cell origin back to (x,y) space
	float Y0 = j-t;
	float x0 = xin-X0; // The x,y distances from the cell origin
	float y0 = yin-Y0;

	// For the 2D case, the simplex shape is an equilateral triangle.
	// Determine which simplex we are in.
	int i1, j1; // Offsets for second (middle) corner of simplex in (i,j) coords
	if(x0>y0) {i1=1; j1=0;} // lower triangle, XY order: (0,0)->(1,0)->(1,1)
	else {i1=0; j1=1;}      // upper triangle, YX order: (0,0)->(0,1)->(1,1)

	// A step of (1,0) in (i,j) means a step of (1-c,-c) in (x,y), and
	// a step of (0,1) in (i,j) means a step of (-c,1-c) in (x,y), where
	// c = (3-sqrt(3))/6

	float x1 = x0 - i1 + G2; // Offsets for middle corner in (x,y) unskewed coords
	float y1 = y0 - j1 + G2;
	float x2 = x0 - 1.0 + 2.0 * G2; // Offsets for last corner in (x,y) unskewed coords
	float y2 = y0 - 1.0 + 2.0 * G2;

	// Work out the hashed gradient indices of the three simplex corners
	int ii = i & 255;
	int jj = j & 255;
	int gi0 = perm[ii+perm[jj]] % 12;
	int gi1 = perm[ii+i1+perm[jj+j1]] % 12;
	int gi2 = perm[ii+1+perm[jj+1]] % 12;

	// Calculate the contribution from the three corners
	float t0 = 0.5 - x0*x0-y0*y0;
	if(t0<0) n0 = 0.0;
	else {
		t0 *= t0;
		n0 = t0 * t0 * (grad3[gi0][0] * x0 + grad3[gi0][1] * y0);  // (x,y) of grad3 used for 2D gradient
	}

	float t1 = 0.5 - x1*x1-y1*y1;
	if(t1<0) n1 = 0.0;
	else {
		t1 *= t1;
		n1 = t1 * t1 * (grad3[gi1][0] * x1 + grad3[gi1][1] * y1);
	}

	float t2 = 0.5 - x2*x2-y2*y2;
	if(t2<0) n2 = 0.0;
	else {
		t2 *= t2;
		n2 = t2 * t2 * (grad3[gi2][0] * x2 + grad3[gi2][1] * y2);
	}

	// Add contributions from each corner to get the final noise value.
	// The result is scaled to return values in the interval [-1,1].
	return 70.0 * (n0 + n1 + n2);
}

void WaterSurface::printStatistics()
{
	std::cout << "Water surface" << std::endl;

	ttDeparture.PrintTime();	ttDeparture.Reset();
	ttPrepSolver.PrintTime();	ttPrepSolver.Reset();
	ttSolver.PrintTime();		ttSolver.Reset();
	ttPostSolver.PrintTime();	ttPostSolver.Reset();
	ttDraw.PrintTime();			ttDraw.Reset();
}