#include "StdAfx.h"
#include "WaterSurfaceCGGPU.h"

WaterSurfaceCGGPU::WaterSurfaceCGGPU(unsigned int surfaceSize, float spacing, float waterLevel)
: WaterSurfaceGPU(surfaceSize, spacing, waterLevel), fbo(surfaceSize, surfaceSize),
  a5vectormultShader(surfaceSize, surfaceSize),
  subtractShader(surfaceSize, surfaceSize),
  innermultShader(surfaceSize, surfaceSize),
  sumreductionShader(surfaceSize, surfaceSize),
  scalarmultShader(surfaceSize, surfaceSize),
  addShader(surfaceSize, surfaceSize)
{
	a5vectormultShader.addVertexSource("default.vert");
	a5vectormultShader.addFragmentSource("a5vecmult.frag");

	subtractShader.addVertexSource("default.vert");
	subtractShader.addFragmentSource("subtract.frag");

	innermultShader.addVertexSource("default.vert");
	innermultShader.addFragmentSource("innermult.frag");

	sumreductionShader.addVertexSource("default.vert");
	sumreductionShader.addFragmentSource("sumreduction.frag");

	scalarmultShader.addVertexSource("default.vert");
	scalarmultShader.addFragmentSource("scalarmult.frag");

	addShader.addVertexSource("default.vert");
	addShader.addFragmentSource("add.frag");

	texA = createTexture(NULL);
	texH = createTexture(NULL);
	texB = createTexture(NULL);
	texR0 = createTexture(NULL);
	texR1 = createTexture(NULL);
	texR01 = createTexture(NULL);
	texRi = createTexture(NULL);
	texPi = createTexture(NULL);
	texAPi = createTexture(NULL);
}

WaterSurfaceCGGPU::~WaterSurfaceCGGPU()
{
	glDeleteTextures(1, &texA);
	glDeleteTextures(1, &texH);
	glDeleteTextures(1, &texB);
	glDeleteTextures(1, &texR0);
	glDeleteTextures(1, &texR1);
	glDeleteTextures(1, &texR01);
	glDeleteTextures(1, &texRi);
	glDeleteTextures(1, &texPi);
	glDeleteTextures(1, &texAPi);
}

void WaterSurfaceCGGPU::solver()
{
	CG(1, 0.01f);
}

void WaterSurfaceCGGPU::CG(int imax, float tolerance)
{
	checkGLErrors("start CG");

	ttOverhead.Start();
	a5vectormultShader.initialize2D(size, size);
	fbo.bind();

	writeTexture(texA, A, NULL, NULL, NULL);
	writeTexture(texH, h, NULL, NULL, NULL);
	writeTexture(texB, b, NULL, NULL, NULL);
	ttOverhead.Stop();

	// do the CG
	float alpha_i;
	float beta_i;
	float d_a_p_i;

	ttShader.Start();
	matrixA5VectorMult(texH, texR0);					// r_i = _b - mat5VecProd(_A, _h);
	subtract(texB, texR0, texRi);
	ttShader.Stop();
	ttOverhead.Start();
	copyTexture(fbo, texRi, texPi);						// p_i = r_i;
	ttOverhead.Stop();

	ttShader.Start();
	for (GLint i = 0; i < imax; i++)
	{
		matrixA5VectorMult(texPi, texAPi);				// a_p_i = mat5VecProd(_A, p_i);
		d_a_p_i = dotProduct(texPi, texAPi);			// d_a_p_i = inner_prod(p_i, a_p_i);
		alpha_i = dotProduct(texRi, texPi) / d_a_p_i;	// alpha_i = inner_prod(r_i,p_i) / d_a_p_i;
		scalarMult(alpha_i, texPi, texR0);				// _h += alpha_i * p_i;
		add(texH, texR0, texH);

		if (i == imax-1) {
			// shortcut out
			break;
		}

		// prepare next iteration
		scalarMult(alpha_i, texAPi, texR0);				// noalias(r_i) -= alpha_i * a_p_i;
		subtract(texRi, texR0, texRi);
		
		if (length(texRi) < tolerance) {				// if (norm_2(r_i) < tolerance) {
			// another shortcut
			break;
		}

		matrixA5VectorMult(texRi, texR0);				// beta_i = - inner_prod(r_i,mat5VecProd(_A,r_i)) / d_a_p_i;
		beta_i = - dotProduct(texRi, texR0) / d_a_p_i;

		scalarMult(beta_i, texPi, texR0);				// p_i = r_i + beta_i * p_i;
		add(texRi, texR0, texPi);
	}
	ttShader.Stop();
	ttOverhead.Start();
	fbo.attach(0, texH);
	fbo.read(0, h, 0);

	fbo.unbind();
	a5vectormultShader.reset();
	ttOverhead.Stop();
}

void WaterSurfaceCGGPU::matrixA5VectorMult(unsigned int texX, unsigned int texResult)
{
	// stores result in x, x must've been allocated with size _nx*_nz
	// b must also be _nx*_nz large, which also must be the number of rows
	// in the _A member.

	if (!fbo.attach(0, texResult))
	{
		return;
	}

	// load and run a5vecmult fragment shader to do inner multiplication
	a5vectormultShader.load();

	a5vectormultShader.setSampler2D("A", GL_TEXTURE0, texA);
	a5vectormultShader.setSampler2D("X", GL_TEXTURE1, texX);
	a5vectormultShader.setUniform1f("Width", (float) size);

	a5vectormultShader.draw2D(0);

	a5vectormultShader.unload();
}

void WaterSurfaceCGGPU::subtract(unsigned int texX, unsigned int texY, unsigned int texResult)
{
	/* --- component subtraction --- */

	if (!fbo.attach(0, texResult))
		return;

	// load and run innermult fragment shader to do inner multiplication
	subtractShader.load();

	subtractShader.setSampler2D("X", GL_TEXTURE0, texX);
	subtractShader.setSampler2D("Y", GL_TEXTURE1, texY);

	subtractShader.draw2D(0);

	subtractShader.unload();
}

float WaterSurfaceCGGPU::dotProduct(GLuint texX, GLuint texY)
{
	/* --- inner multiplication --- */

	if (!fbo.attach(0, texR0))
		return 0.0f;

	// load and run innermult fragment shader to do inner multiplication
	innermultShader.load();

	innermultShader.setSampler2D("X", GL_TEXTURE0, texX);
	innermultShader.setSampler2D("Y", GL_TEXTURE1, texY);
	innermultShader.draw2D(0);

	innermultShader.unload();

	// R0 contain result of inner multiplication

	/* --- sum-reduction with add operation --- */

	// setup ping pong rendering
	GLenum attachmentPoints[] = {
		0,
		1
	};
	GLuint textures[] = {
		texR0,
		texR1
	};
	int readTexture = 0;
	int writeTexture = 1;

	if (!fbo.attach(attachmentPoints[readTexture], textures[readTexture]))
		return 0.0f;
	if (!fbo.attach(attachmentPoints[writeTexture], textures[writeTexture]))
		return 0.0f;

	// perform sum reduction
	sumreductionShader.load();

	int halfsize = size / 2;
	float texelDist = 1.0f / (float)size;

	while (halfsize > 0)
	{
		sumreductionShader.setSampler2D("X", GL_TEXTURE0, textures[readTexture]);
		sumreductionShader.setUniform1f("TexelDist", (float) texelDist);
		sumreductionShader.draw2D(attachmentPoints[writeTexture]);
		
		// swap
		int temp = readTexture;
		readTexture = writeTexture;
		writeTexture = temp;

		halfsize /= 2;
		texelDist *= 2.0f;
	}
	sumreductionShader.unload();

	// read out result
	float result[4];
	glReadBuffer(GL_COLOR_ATTACHMENT0_EXT + attachmentPoints[readTexture]); // readTexture since it's been swapped
	glReadPixels(0, 0, 1, 1, GL_RGBA, GL_FLOAT, result); // read only the first value

	return result[0];
}

void WaterSurfaceCGGPU::scalarMult(float scalar, unsigned int texX, unsigned int texResult)
{
	/* --- scalar multiplication --- */

	if (!fbo.attach(0, texResult))
		return;

	scalarmultShader.load();

	scalarmultShader.setSampler2D("X", GL_TEXTURE0, texX);
	scalarmultShader.setUniform1f("S", scalar);
	scalarmultShader.draw2D(0);

	scalarmultShader.unload();
}

void WaterSurfaceCGGPU::add(unsigned int texX, unsigned int texY, unsigned int texResult)
{
	/* --- component addition --- */

	if (!fbo.attach(0, texResult))
		return;

	// load and run innermult fragment shader to do inner multiplication
	addShader.load();

	addShader.setSampler2D("X", GL_TEXTURE0, texX);
	addShader.setSampler2D("Y", GL_TEXTURE1, texY);
	addShader.draw2D(0);

	addShader.unload();
}

float WaterSurfaceCGGPU::length(GLuint texX)
{
	return sqrt(dotProduct(texX, texX));
}

