#include "StdAfx.h"
#include "WaterSurfaceFullGPU.h"

WaterSurfaceFullGPU::WaterSurfaceFullGPU(unsigned int surfaceSize, float spacing, float waterLevel)
: WaterSurfaceGPU(surfaceSize, spacing, waterLevel), fbo(surfaceSize, surfaceSize),
  updateTextures(true)
{
	departureShader.addVertexSource("default.vert");
	departureShader.addFragmentSource("fullDeparture.frag");

	prepareJacobiShader.addVertexSource("default.vert");
	prepareJacobiShader.addFragmentSource("fullPrepareJacobi.frag");

	jacobiShader.addVertexSource("default.vert");
	jacobiShader.addFragmentSource("fullJacobi.frag");

	postJacobiShader.addVertexSource("default.vert");
	postJacobiShader.addFragmentSource("fullPostJacobi.frag");

	texH = createTexture(NULL);
	texAltH = createTexture(NULL);
	texUV = createTexture(NULL);
	texTildeHUV = createTexture(NULL);
	texAB = createTexture(NULL);
}

WaterSurfaceFullGPU::~WaterSurfaceFullGPU()
{
	glDeleteTextures(1, &texH);
	glDeleteTextures(1, &texAltH);
	glDeleteTextures(1, &texUV);
	glDeleteTextures(1, &texTildeHUV);
	glDeleteTextures(1, &texAB);
}

void WaterSurfaceFullGPU::update(int ticks)
{
	if (updateTextures)
	{
		writeTexture(texH, h, NULL, NULL, NULL);
		writeTexture(texUV, u, v, NULL, NULL);
		updateTextures = false;
	}

	WaterSurfaceGPU::update(ticks);
	updateTextures = true;
}

void WaterSurfaceFullGPU::computeDeparture()
{
	//WaterSurface::computeDeparture();
	//writeTexture(texTildeHUV, tildeH, tildeU, tildeV, NULL);
	//printTable("CPU tildeH", tildeH);
	//printTable("CPU tildeU", tildeU);
	//printTable("CPU tildeV", tildeV);
	
	checkGLErrors("start departure");

	departureShader.initialize2D(size, size);

	// attach write-texture to color attachment
	fbo.bind();
	if (!fbo.attach(0, texTildeHUV))
		return;
	checkGLErrors("setup fbo");

	// load and run deppoint shader
	departureShader.load();
	departureShader.setSampler2D("texH", GL_TEXTURE0, texH);
	departureShader.setSampler2D("texUV", GL_TEXTURE1, texUV);
	departureShader.setUniform1i("size", size);
	departureShader.setUniform1f("dT", dT);
	departureShader.setUniform1f("dS", dS);

	departureShader.draw2D(0);

	departureShader.unload();
	checkGLErrors("departure shader");

	// until post jacobi is implemented
	fbo.read(0, tildeH, tildeU, tildeV, NULL);
	//printTable("GPU tildeH", tildeH);
	//printTable("GPU tildeU", tildeU);
	//printTable("GPU tildeV", tildeV);

	fbo.unbind();
	departureShader.reset();

	checkGLErrors("end departure");
}

void WaterSurfaceFullGPU::prepareSolver()
{
	//WaterSurface::prepareSolver();
	//writeTexture(texAB, A, b, NULL, NULL);
	//printTable("CPU A", A);
	//printTable("CPU b", b);

	checkGLErrors("start prepare");

	prepareJacobiShader.initialize2D(size, size);

	// attach write-texture to color attachment
	fbo.bind();
	if (!fbo.attach(0, texAB))
		return;
	checkGLErrors("setup fbo");

	// load and run deppoint shader
	prepareJacobiShader.load();
	prepareJacobiShader.setSampler2D("texH", GL_TEXTURE0, texH);
	prepareJacobiShader.setSampler2D("texTildeHUV", GL_TEXTURE1, texTildeHUV);
	prepareJacobiShader.setUniform1i("size", size);
	prepareJacobiShader.setUniform1f("dT", dT);
	prepareJacobiShader.setUniform1f("dS", dS);
	prepareJacobiShader.setUniform1f("g", g);

	prepareJacobiShader.draw2D(0);

	prepareJacobiShader.unload();
	checkGLErrors("prepare");

	fbo.read(0, A, b, NULL, NULL);
	//printTable("GPU A", A);
	//printTable("GPU b", b);

	fbo.unbind();
	prepareJacobiShader.reset();

	checkGLErrors("end prepare");
}

void WaterSurfaceFullGPU::solver()
{
	//jacobiCPU(5, 0.1f);
	//printTable("CPU h", h);
	jacobi(5);
	//printTable("GPU h", h);
}

void WaterSurfaceFullGPU::jacobi(int imax)
{
	checkGLErrors("start jacobi");

	/* --- jacobi shader --- */

	jacobiShader.initialize2D(size, size);

	// setup texture swapping
	GLenum attachmentPoints[] = {
		0, 
		1
	};
	unsigned int textures[] = {
		texH,
		texAltH
	};
	unsigned int readTex = 0;
	unsigned int writeTex = 1;

	// attach textures to color attachments
	fbo.bind();
	fbo.attach(attachmentPoints[readTex], textures[readTex]);
	fbo.attach(attachmentPoints[writeTex], textures[writeTex]);
	checkGLErrors("setup fbo");

	// load and run jacobi shader
	jacobiShader.load();
	jacobiShader.setSampler2D("texAB", GL_TEXTURE1, texAB);

	// ensure that the final jacobi iteration writes to texH
	if (imax % 2 > 0) imax++;

	for (int i = 0; i < imax; i++)
	{
		jacobiShader.setSampler2D("texH", GL_TEXTURE0, textures[readTex]);
		//jacobiShader.setSampler2D("texAB", GL_TEXTURE1, texAB);

		jacobiShader.draw2D(attachmentPoints[writeTex]);

		// swap textures
		int temp = readTex;
		readTex = writeTex;
		writeTex = temp;
	}

	jacobiShader.unload();
	checkGLErrors("jacobi shader");

	// read the result
	fbo.read(attachmentPoints[readTex], h, 0);
	checkGLErrors("readFBO");

	// restore current viewport
	//jacobiFBO.unattach(attachmentPoints[readTex]);
	//jacobiFBO.unattach(attachmentPoints[writeTex]);
	fbo.unbind();

	jacobiShader.reset();
	glActiveTexture(GL_TEXTURE0);

	checkGLErrors("end jacobi");

	updateTextures = true; // since post jacobi isn't implemented yet
}

void WaterSurfaceFullGPU::jacobiCPU(int imax, float tolerance)
{
	float *new_h = new float[size_2];
	int iter_num = 0;
	float sqrtolr = tolerance*tolerance;
	float sqrdiff;

	do {
		for (unsigned int i = 0; i < size; i++) {
			for (unsigned int j = 0; j < size; j++) {
				int pos = i*size + j;
				//float delta = 0.0f;

				/*if (j > 0)		delta -= h[pos];
				if (i > 0)		delta -= h[pos];
				if (i < size-1)	delta -= h[pos];
				if (j < size-1)	delta -= h[pos];*/

				new_h[pos] = (b[pos] + 4.0f*h[pos]) / A[pos];
			}
		}

		// check for convergence
		sqrdiff = sqrdist(new_h, h);

		// copy the new height map to the height map
		memcpy(h, new_h, sizeof(float) * size_2);

		iter_num++;
	} while ((iter_num < imax)); // && (sqrdiff >= sqrtolr)); // TODO slutte hvis diff blir strre?
	//std::cout << "Iterations: " << iter_num << std::endl;
	delete[] new_h;
}

float WaterSurfaceFullGPU::sqrdist(float *x, float *y) 
{
	float sum = 0.0;

	for (unsigned int i = 0; i < size_2; i++) {
		float value = x[i] - y[i];
		sum += value * value;
	}
	return sum;
}

