#include "StdAfx.h"
#include "WaterSurfaceJacobiGPU.h"

WaterSurfaceJacobiGPU::WaterSurfaceJacobiGPU(unsigned int surfaceSize, float spacing, float waterLevel)
: WaterSurfaceGPU(surfaceSize, spacing, waterLevel), jacobiFBO(surfaceSize, surfaceSize)
{
	jacobiShader.addVertexSource("default.vert");
	jacobiShader.addFragmentSource("jacobi.frag");

	jacobiTex1 = createTexture(NULL);
	jacobiTex2 = createTexture(NULL);
}

WaterSurfaceJacobiGPU::~WaterSurfaceJacobiGPU()
{
	glDeleteTextures(1, &jacobiTex1);
	glDeleteTextures(1, &jacobiTex2);
}

void WaterSurfaceJacobiGPU::solver()
{
	jacobi(5, dT/100000.0f);
}

void WaterSurfaceJacobiGPU::jacobi(int imax, float tolerance)
{
	checkGLErrors("start jacobi");

	/* --- jacobi shader --- */

	ttOverhead.Start();
	jacobiShader.initialize2D(size, size);
	jacobiFBO.bind();

	// write to texture first read
	writeTexture(jacobiTex1, A, h, b, NULL);
	checkGLErrors("write texture");

	// setup texture swapping
	GLenum attachmentPoints[] = {
		0, 
		1
	};
	unsigned int textures[] = {
		jacobiTex1,
		jacobiTex2
	};
	unsigned int readTex = 0;
	unsigned int writeTex = 1;

	// attach textures to color attachments
	jacobiFBO.attach(attachmentPoints[readTex], textures[readTex]);
	jacobiFBO.attach(attachmentPoints[writeTex], textures[writeTex]);
	checkGLErrors("setup fbo");

	// load and run jacobi shader
	jacobiShader.load();
	ttOverhead.Stop();

	for (int i = 0; i < imax; i++)
	{
		ttOverhead.Start();
		jacobiShader.setSampler2D("tex", GL_TEXTURE0, textures[readTex]);
		ttOverhead.Stop();

		ttShader.Start();
		jacobiShader.draw2D(attachmentPoints[writeTex]);
		ttShader.Stop();

		// swap textures
		int temp = readTex;
		readTex = writeTex;
		writeTex = temp;
	}

	ttOverhead.Start();
	jacobiShader.unload();
	checkGLErrors("jacobi shader");

	glBindTexture(GL_TEXTURE_2D, 0);
	
	// read the result
	jacobiFBO.read(attachmentPoints[readTex], h, 1);
	checkGLErrors("readFBO");

	// restore current viewport
	jacobiFBO.unbind();
	jacobiShader.reset();

	ttOverhead.Stop();
	checkGLErrors("end jacobi");
}

WaterSurfaceJacobiPackedGPU::WaterSurfaceJacobiPackedGPU(unsigned int surfaceSize, float spacing, float waterLevel)
: WaterSurfaceGPU(surfaceSize, spacing, waterLevel), jacobiFBO(surfaceSize/2, surfaceSize/2)
{
	jacobiShader.addVertexSource("default.vert");
	jacobiShader.addFragmentSource("jacobipacked.frag");

	texA = createPackedTexture(NULL);
	texh1 = createPackedTexture(NULL);
	texh2 = createPackedTexture(NULL);
	texb = createPackedTexture(NULL);
}

WaterSurfaceJacobiPackedGPU::~WaterSurfaceJacobiPackedGPU()
{
	glDeleteTextures(1, &texA);
	glDeleteTextures(1, &texh1);
	glDeleteTextures(1, &texh2);
	glDeleteTextures(1, &texb);
}

void WaterSurfaceJacobiPackedGPU::solver()
{
	jacobi(5, dT/100000.0f);
}

void WaterSurfaceJacobiPackedGPU::jacobi(int imax, float tolerance)
{
	checkGLErrors("start jacobi");

	/* --- jacobi shader --- */

	ttOverhead.Start();

	jacobiShader.initialize2D(size/2, size/2);
	jacobiFBO.bind();

	writePackedTexture(texA, A);
	writePackedTexture(texh1, h);
	writePackedTexture(texb, b);
	checkGLErrors("write texture");

	// setup texture swapping
	GLenum attachmentPoints[] = {
		0, 
		1
	};
	unsigned int textures[] = {
		texh1,
		texh2
	};
	unsigned int readTex = 0;
	unsigned int writeTex = 1;

	// attach textures to color attachments
	jacobiFBO.attach(attachmentPoints[readTex], textures[readTex]);
	jacobiFBO.attach(attachmentPoints[writeTex], textures[writeTex]);
	checkGLErrors("setup fbo");

	// load and run jacobi shader
	jacobiShader.load();

	jacobiShader.setSampler2D("texA", GL_TEXTURE0, texA);
	jacobiShader.setSampler2D("texb", GL_TEXTURE2, texb);
	ttOverhead.Stop();

	for (int i = 0; i < imax; i++)
	{
		ttOverhead.Start();
		jacobiShader.setSampler2D("texh", GL_TEXTURE1, textures[readTex]);
		ttOverhead.Stop();

		ttShader.Start();
		jacobiShader.draw2D(attachmentPoints[writeTex]);
		ttShader.Stop();

		// swap textures
		int temp = readTex;
		readTex = writeTex;
		writeTex = temp;
	}

	ttOverhead.Start();
	jacobiShader.unload();
	checkGLErrors("jacobi shader");

	glActiveTexture(GL_TEXTURE0);
	glBindTexture(GL_TEXTURE_2D, 0);
	
	// read the result
	jacobiFBO.readPacked(attachmentPoints[readTex], h);
	checkGLErrors("readFBO");

	// restore current viewport
	jacobiFBO.unbind();
	jacobiShader.reset();

	checkGLErrors("end jacobi");
	ttOverhead.Stop();
}
